"""Utilities."""
from functools import wraps
import re
from typing import Callable, List, Optional, TypeVar, Union

from .data import (
    all_classes, all_slots,
)


def pascal_to_snake(s: str, sep: str = "_") -> str:
    """Convert Pascal case to snake case.

    Assumes that
    a) all words are either all-lowercase or all-uppercase
    b) all 1-letter words are lowercase
    c) there are no adjacent 1-letter words
    d) there are no adjacent uppercase words

    Examples:
    PhenotypicFeature -> phenotypic_feature
    RNAProduct -> RNA_product
    FeedACamel -> feed_a_camel
    
    Optionally specify `sep` (default "_").
    """
    # add an underscore before each capital letter
    underscored = re.sub(
        r"(?<!^)(?=[A-Z])",
        sep,
        s,
    )
    # collapse any adjacent one-letter words
    collapsed = re.sub(
        r"(?<![a-zA-Z])[A-Z](?:_[A-Z](?=$|_))+",
        lambda match: match.group(0).replace("_", ""),
        underscored,
    )
    # lower-case any words containing only one uppercase letter
    lowercased = re.sub(
        r"(?<![A-Z])[A-Z](?![A-Z])",
        lambda match: match.group(0).lower(),
        collapsed,
    )
    return lowercased


def snake_to_pascal(s: str, sep: str = "_") -> str:
    """Convert snake case to Pascal case.

    This is the inverse of pascal_to_snake() when its assumptions
    are true.
    
    Optionally specify `sep` (default "_").
    """
    return re.sub(
        fr"(?:^|{sep})([a-zA-Z])",
        lambda match: match.group(1).upper(),
        s
    )


def guess_casing(s: str) -> str:
    """Guess snake case or Pascal case."""
    if "_" in s:
        return "snake"
    if any(c.isupper() for c in s):
        return "pascal"
    return "snake"


def normalize(s: str) -> str:
    """Normalize string input."""
    if s.startswith("biolink:"):
        s = s[8:]
    if "_" in s:
        # it's snake case
        return s.replace("_", " ")
    if " " in s:
        return s
    return pascal_to_snake(s, " ")


T = TypeVar("T")


def listify(func: Callable) -> Callable:
    """Expand function to take list of arguments."""
    @wraps(func)
    def wrapper(arg: Union[T, List[T]], **kwargs) -> Union[T, List[T]]:
        """Apply function to each element in list."""
        if isinstance(arg, list):
            return [
                func(el, **kwargs)
                for el in arg
            ]
        else:
            return func(arg, **kwargs)
    return wrapper


@listify
def format(s: str, case: Optional[str] = None, **kwargs) -> str:
    """Format space-case string as biolink CURIE."""
    if isinstance(case, str) and case.lower() == "pascal":
            return "biolink:" + snake_to_pascal(s, " ")
    elif isinstance(case, str) and case.lower() == "snake":
            return "biolink:" + s.replace(" ", "_")
    else:
        return "biolink:" + s


def with_formatting():
    """Add format conversions to method."""
    def decorator(func: Callable) -> Callable:
        """Generate decorator."""
        @wraps(func)
        def wrapper(self, s: str, *args, formatted=False, **kwargs):
            """Wrap in format conversions."""
            case = guess_casing(s)
            normalized = normalize(s)
            output: Union[str, List[str]] = func(self, normalized, *args, **kwargs)
            if formatted:
                if normalized in all_classes:
                    output = format(output, case="pascal")
                elif normalized in all_slots:
                    output = format(output, case="snake")
                else:
                    output = format(output, case=case)
            return output
        return wrapper
    return decorator



"""Biolink model data."""
import json
from typing import Dict, List, Optional

import importlib.resources as pkg_resources

from . import _data

with pkg_resources.open_text(_data, "all_classes.json") as stream:
    all_classes: List[str] = json.load(stream)
with pkg_resources.open_text(_data, "all_elements.json") as stream:
    all_elements: List[str] = json.load(stream)
with pkg_resources.open_text(_data, "all_slots.json") as stream:
    all_slots: List[str] = json.load(stream)
with pkg_resources.open_text(_data, "all_types.json") as stream:
    all_types: List[str] = json.load(stream)
with pkg_resources.open_text(_data, "alias_ancestors.json") as stream:
    alias_ancestors: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "basic_ancestors.json") as stream:
    basic_ancestors: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "mixin_ancestors.json") as stream:
    mixin_ancestors: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "alias_mixin_ancestors.json") as stream:
    alias_mixin_ancestors: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "alias_descendants.json") as stream:
    alias_descendants: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "basic_descendants.json") as stream:
    basic_descendants: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "mixin_descendants.json") as stream:
    mixin_descendants: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "alias_mixin_descendants.json") as stream:
    alias_mixin_descendants: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "alias_children.json") as stream:
    alias_children: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "basic_children.json") as stream:
    basic_children: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "mixin_children.json") as stream:
    mixin_children: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "alias_mixin_children.json") as stream:
    alias_mixin_children: Dict[str, List[str]] = json.load(stream)
with pkg_resources.open_text(_data, "parent.json") as stream:
    parent: Dict[str, Optional[str]] = json.load(stream)
with pkg_resources.open_text(_data, "element.json") as stream:
    element: Dict[str, Dict] = json.load(stream)



"""BMT basics."""
from typing import Dict, List, Optional

from .data import (
    all_classes, all_elements, all_slots, all_types,
    alias_ancestors, basic_ancestors, mixin_ancestors, alias_mixin_ancestors,
    alias_descendants, basic_descendants, mixin_descendants, alias_mixin_descendants,
    alias_children, basic_children, mixin_children, alias_mixin_children,
    parent, element,
)
from .util import with_formatting


class Toolkit():
    """Biolink model toolkit - lite!"""

    def __init__(self, schema=None):
        """Initialize."""
        if schema is not None:
            raise ValueError("bmt-lite does not support the `schema` argument. The biolink model version is dictated by the library flavor you installed.")

    def get_all_classes(self) -> List[str]:
        """Get all classes."""
        return all_classes

    def get_all_slots(self) -> List[str]:
        """Get all slots."""
        return all_slots

    def get_all_types(self) -> List[str]:
        """Get all types."""
        return all_types

    def get_all_elements(self) -> List[str]:
        """Get all elements."""
        return all_elements

    @with_formatting()
    def get_ancestors(
        self,
        name: str,
        reflexive: bool = True,
        mixin: bool = True,
        alias: bool = False,
    ) -> List[str]:
        """Get ancestors."""
        _ancestors = basic_ancestors.get(name, [])
        if mixin:
            _ancestors += mixin_ancestors.get(name, [])
        if alias:
            _ancestors += alias_ancestors.get(name, [])
        if mixin and alias:
            _ancestors += alias_mixin_ancestors.get(name, [])
        if reflexive:
            return _ancestors + [name]
        else:
            return _ancestors

    @with_formatting()
    def get_descendants(
        self,
        name: str,
        reflexive: bool = True,
        mixin: bool = True,
        alias: bool = False,
    ) -> List[str]:
        """Get descendants."""
        _descendants = basic_descendants.get(name, [])
        if mixin:
            _descendants += mixin_descendants.get(name, [])
        if alias:
            _descendants += alias_descendants.get(name, [])
        if mixin and alias:
            _descendants += alias_mixin_descendants.get(name, [])
        if reflexive:
            return _descendants + [name]
        else:
            return _descendants

    @with_formatting()
    def get_children(
        self,
        name: str,
        mixin: bool = True,
        alias: bool = False,
    ) -> List[str]:
        """Get children."""
        _children = basic_children.get(name, [])
        if mixin:
            _children += mixin_children.get(name, [])
        if alias:
            _children += alias_children.get(name, [])
        if mixin and alias:
            _children += alias_mixin_children.get(name, [])
        return _children

    @with_formatting()
    def get_parent(
        self,
        name: str,
    ) -> Optional[str]:
        """Get parent."""
        return parent.get(name, None)

    @with_formatting()
    def get_element(
        self,
        name: str,
    ) -> Optional["Element"]:
        """Get element."""
        if name in all_classes:
            return ClassDefinition(name, **element.get(name, dict()))
        elif name in all_slots:
            return SlotDefinition(name, **element.get(name, dict()))
        return None


class AttrDict(dict):
    """https://stackoverflow.com/a/14620633"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class Element(AttrDict):
    """Biolink model element."""

    def __init__(self, name: str):
        """Initialize."""
        super().__init__()
        self.name: str = name


class SlotDefinition(Element):
    """Slot definition."""

    def __init__(
        self,
        name: str,
        symmetric: bool = False,
        inverse: Optional[str] = None,
        annotations: Dict[str, bool] = dict(),
        slot_uri: Optional[str] = None,
        range: Optional[str] = None,
        **kwargs,
    ):
        """Initialize."""
        super().__init__(name)
        self.symmetric: bool = symmetric
        self.inverse: Optional[str] = inverse
        self.annotations: Dict[str, bool] = annotations
        self.slot_uri: Optional[str] = slot_uri
        self.range: Optional[str] = range


class ClassDefinition(Element):
    """Class definition."""

    def __init__(
        self,
        name: str,
        id_prefixes: List[str],
        mixins: List[str],
    ):
        """Initialize."""
        super().__init__(name)
        self.id_prefixes: List[str] = id_prefixes
        self.mixins: List[str] = mixins



"""Set up bmt-lite package."""
import json
from pathlib import Path
import re
from setuptools import setup
import sys

stash = sys.path.pop(0)  # avoid trying to import the local bmt
from bmt import Toolkit
sys.path = [stash] + sys.path  # restore the path
import httpx

FILEPATH = Path(__file__).parent

DATAPATH = Path("bmt/_data")
DATAPATH.mkdir(exist_ok=True)  # create data dir
(DATAPATH / "__init__.py").touch(exist_ok=True)  # make data path a module

response = httpx.get("https://api.github.com/repos/biolink/biolink-model/releases")
releases = response.json()
versions = [
    release["tag_name"]
    for release in releases
]


def build(version: str):
    """Build BMT data."""
    if version in versions:
        BMT = Toolkit(
            schema=f"https://raw.githubusercontent.com/biolink/biolink-model/{version}/biolink-model.yaml",
        )
    elif "v" + version in versions:
        BMT = Toolkit(
            schema=f"https://raw.githubusercontent.com/biolink/biolink-model/v{version}/biolink-model.yaml",
        )
    elif version.removeprefix("v") in versions:
        BMT = Toolkit(
            schema=f"https://raw.githubusercontent.com/biolink/biolink-model/{version[1:]}/biolink-model.yaml",
        )

    # get_all_classes()
    classes = BMT.get_all_classes()
    with open(DATAPATH / "all_classes.json", "w") as stream:
        json.dump(classes, stream)

    # get_all_slots()
    slots = BMT.get_all_slots()
    with open(DATAPATH / "all_slots.json", "w") as stream:
        json.dump(slots, stream)

    # get_all_types()
    types = BMT.get_all_types()
    with open(DATAPATH / "all_types.json", "w") as stream:
        json.dump(types, stream)

    # get_all_elements()
    elements = classes + slots  + types
    with open(DATAPATH / "all_elements.json", "w") as stream:
        json.dump(elements, stream)

    # get_ancestors()
    basic_ancestors = {
        element: BMT.get_ancestors(
            element,
            reflexive=False,
            mixin=False,
        )
        for element in elements
    }
    alias_ancestors = {
        element: [
            alias
            for ancestor in basics
            if (el := BMT.get_element(ancestor)) is not None and (aliases := el.aliases) is not None
            for alias in aliases
        ]
        for element, basics in basic_ancestors.items()
    }
    mixin_ancestors = {
        element: [
            ancestor
            for ancestor in BMT.get_ancestors(
                element,
                reflexive=False,
                mixin=True,
            )
            if ancestor not in basic_ancestors[element]
        ]
        for element in elements
    }
    alias_mixin_ancestors = {
        element: [
            alias
            for ancestor in mixins
            if (el := BMT.get_element(ancestor)) is not None and (aliases := el.aliases) is not None
            for alias in aliases
        ]
        for element, mixins in mixin_ancestors.items()
    }
    with open(DATAPATH / "basic_ancestors.json", "w") as stream:
        json.dump(basic_ancestors, stream)
    with open(DATAPATH / "alias_ancestors.json", "w") as stream:
        json.dump(alias_ancestors, stream)
    with open(DATAPATH / "mixin_ancestors.json", "w") as stream:
        json.dump(mixin_ancestors, stream)
    with open(DATAPATH / "alias_mixin_ancestors.json", "w") as stream:
        json.dump(alias_mixin_ancestors, stream)

    # get_descendants()
    basic_descendants = {
        element: BMT.get_descendants(
            element,
            reflexive=False,
            mixin=False,
        )
        for element in elements
    }
    alias_descendants = {
        element: [
            alias
            for descendant in basics
            if (el := BMT.get_element(descendant)) is not None and (aliases := el.aliases) is not None
            for alias in aliases
        ]
        for element, basics in basic_descendants.items()
    }
    mixin_descendants = {
        element: [
            descendant
            for descendant in BMT.get_descendants(
                element,
                reflexive=False,
                mixin=True,
            )
            if descendant not in basic_descendants[element]
        ]
        for element in elements
    }
    alias_mixin_descendants = {
        element: [
            alias
            for descendant in mixins
            if (el := BMT.get_element(descendant)) is not None and (aliases := el.aliases) is not None
            for alias in aliases
        ]
        for element, mixins in mixin_descendants.items()
    }
    with open(DATAPATH / "basic_descendants.json", "w") as stream:
        json.dump(basic_descendants, stream)
    with open(DATAPATH / "alias_descendants.json", "w") as stream:
        json.dump(alias_descendants, stream)
    with open(DATAPATH / "mixin_descendants.json", "w") as stream:
        json.dump(mixin_descendants, stream)
    with open(DATAPATH / "alias_mixin_descendants.json", "w") as stream:
        json.dump(alias_mixin_descendants, stream)

    # get_children()
    basic_children = {
        element: BMT.get_children(
            element,
            mixin=False,
        )
        for element in elements
    }
    alias_children = {
        element: [
            alias
            for _child in basics
            if (el := BMT.get_element(_child)) is not None and (aliases := el.aliases) is not None
            for alias in aliases
        ]
        for element, basics in basic_children.items()
    }
    mixin_children = {
        element: [
            child
            for child in BMT.get_children(
                element,
                mixin=True,
            )
            if child not in basic_children[element]
        ]
        for element in elements
    }
    alias_mixin_children = {
        element: [
            alias
            for _child in mixins
            if (el := BMT.get_element(_child)) is not None and (aliases := el.aliases) is not None
            for alias in aliases
        ]
        for element, mixins in mixin_children.items()
    }
    with open(DATAPATH / "basic_children.json", "w") as stream:
        json.dump(basic_children, stream)
    with open(DATAPATH / "alias_children.json", "w") as stream:
        json.dump(alias_children, stream)
    with open(DATAPATH / "mixin_children.json", "w") as stream:
        json.dump(mixin_children, stream)
    with open(DATAPATH / "alias_mixin_children.json", "w") as stream:
        json.dump(alias_mixin_children, stream)

    # get_parent()
    parent = {
        element: BMT.get_parent(element)
        for element in elements
    }
    with open(DATAPATH / "parent.json", "w") as stream:
        json.dump(parent, stream)

    # get_element()
    element = dict(
        **{
            class_: {
                "id_prefixes": el.id_prefixes,
                "mixins": el.mixins,
            }
            for class_ in classes
            if (el := BMT.get_element(class_)) is not None
        },
        **{
            slot: {
                "symmetric": el.symmetric,
                "inverse": el.inverse,
                "annotations": {
                    tag: annotation.value.lower() == "true"
                    for tag, annotation in el.annotations.items()
                },
                "slot_uri": el.slot_uri,
                "range": el.range,
            }
            for slot in slots
            if (el := BMT.get_element(slot)) is not None
        }
    )
    with open(DATAPATH / "element.json", "w") as stream:
        json.dump(element, stream)


with open("README.md", "r") as stream:
    long_description = stream.read()

try:
    idx = next(
        idx for idx, arg in enumerate(sys.argv)
        if (match := re.fullmatch(r"--v\d+\.\d+\.\d+", arg)) is not None
    )
except StopIteration:
    print("ERROR: Specify a biolink-model version using the '--vX.Y.Z' argument")
    exit()
version = sys.argv.pop(idx)[3:]
build(version)

setup(
    name=f"bmt-lite-{version}",
    version="2.2.0",
    author="Patrick Wang",
    author_email="patrick@covar.com",
    url="https://github.com/patrickkwang/bmt-lite",
    description="A zero-dependency near-clone of common bmt capabilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["bmt", "bmt._data"],
    package_data={"bmt._data": ["*.json"]},
    include_package_data=True,
    install_requires=[],
    zip_safe=False,
    license="MIT",
    python_requires=">=3.7",
)
